#pragma comment(linker, "/version:1.4") // MUST be in the form of major.minor

#include <windows.h>
#include <shlwapi.h>
#include <stdio.h>
#include <tchar.h>
#include "libs\NtInformationFile.h"
#include "libs\SimpleString.h"
#include "libs\WinHash.h"

#ifndef READ_BUFFER_SIZE
#define READ_BUFFER_SIZE 0x10000
#endif

#define CREATION_ATTRIBUTE_MASK (FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_ARCHIVE | FILE_ATTRIBUTE_NORMAL | FILE_ATTRIBUTE_ENCRYPTED)

typedef struct {
	PVOID pvBuffer;            // file buffer (allocated only once)
	DWORD dwAttributesMask;    // mask of file attributes to preserve
	BOOLEAN fVerbose;          // print status?
	BOOLEAN fPreserveExisting; // refuse to overwrite existing file?
	BOOLEAN fDiscardStreams;   // discard NTFS streams?
	UINT8 fHashAlgorithm;      // hash algorithm
	FILE *stream;              // output stream
	SIZE_T cchTargetPrefix;    // length of prefix portion of the target path
	UINT cSuccessFiles;        // files copied
	UINT cSuccessDirs;         // directories processed
	UINT cErrors;              // errors encountered
	UINT cSkippedFiles;        // files skipped (if fPreserveExisting)
} WORKCTX, *PWORKCTX;

typedef struct {
	DWORD    dwFileAttributes;
	FILETIME ftCreationTime;
	FILETIME ftLastAccessTime;
	FILETIME ftLastWriteTime;
} FILEINFO, *PFILEINFO;

enum {
	TYPE_DNE = 0,
	TYPE_FILE,
	TYPE_DIR,
	TYPE_WILD
};

typedef union {
	UINT16 united;
	struct {
		UINT8 type;
		UINT8 fSlash;
	};
} ITEMTYPE, *PITEMTYPE;

VOID WINAPI CloneFile( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath, PFILEINFO pfi );
VOID WINAPI CloneDirectory( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath, PFILEINFO pfi );
VOID WINAPI CloneFilesIntoDirectory( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath, BOOLEAN fSourceIsDirectoryRoot );
VOID WINAPI PrintHash( PVOID pvResult, PCTSTR pszPath, SIZE_T cdwResult );
UINT16 __fastcall DetermineItemType( PTSTR pszPath );
VOID __fastcall SetSparseAndCompression( HANDLE hFile, DWORD dwAttributes );
VOID __fastcall SetTimeAndAttributes( HANDLE hFile, PFILEINFO pfi );
VOID __fastcall PrintError( PCTSTR pszPath, DWORD dwErrorCode );
__forceinline BOOLEAN IsFlag( PCTSTR pszArg );
__forceinline BOOLEAN CheckFlagI( PCTSTR pszArg, TCHAR ch );
__forceinline BOOLEAN IsSpecialDirectoryName( PCTSTR pszPath );
__forceinline BOOLEAN IsValidDirectory( PCTSTR pszPath );

UINT __fastcall EntryMain( UINT argc, PTSTR *argv )
{
	UINT iSourceStart = 0;
	UINT iSourceEnd;
	PTSTR pszTargetPath = NULL;
	BOOLEAN fShowUsage = FALSE;

	// Parse command line parameters

	BOOLEAN fFlagsDone = FALSE;
	UINT i;

	WORKCTX ctx;
	ZeroMemory(&ctx, sizeof(ctx));
	ctx.dwAttributesMask = ~0;
	ctx.stream = stdout;

	for (i = 1; i < argc; ++i)
	{
		// Check hashing flags, which are multicharacter strings

		if (!fFlagsDone && (argv[i][0] | 0x02) == TEXT('/'))
		{
			UINT8 fNewHashAlgorithm = 0;

			if (StrCmpI(argv[i] + 1, TEXT("crc32")) == 0)
				fNewHashAlgorithm = WHEX_CHECKCRC32;
			else if (StrCmpI(argv[i] + 1, TEXT("md4")) == 0)
				fNewHashAlgorithm = WHEX_CHECKMD4;
			else if (StrCmpI(argv[i] + 1, TEXT("md5")) == 0)
				fNewHashAlgorithm = WHEX_CHECKMD5;
			else if (StrCmpI(argv[i] + 1, TEXT("sha1")) == 0)
				fNewHashAlgorithm = WHEX_CHECKSHA1;
			else if (StrCmpI(argv[i] + 1, TEXT("ed2k")) == 0)
				fNewHashAlgorithm = WHEX_CHECKED2K;

			if (fNewHashAlgorithm)
			{
				if (ctx.fHashAlgorithm)
					fShowUsage = TRUE;
				else
					ctx.fHashAlgorithm = fNewHashAlgorithm;

				continue;
			}
		}

		// Check all other flags and set source(s)/destination

		if (fFlagsDone || !IsFlag(argv[i]))
		{
			fFlagsDone = TRUE;

			if (iSourceStart == 0)
				iSourceStart = i;
			else
			{
				iSourceEnd = i;
				pszTargetPath = argv[i];
			}
		}
		else if (!ctx.fVerbose && CheckFlagI(argv[i], TEXT('v')))
			ctx.fVerbose = TRUE;
		else if (!ctx.fPreserveExisting && CheckFlagI(argv[i], TEXT('p')))
			ctx.fPreserveExisting = TRUE;
		else if (!ctx.fDiscardStreams && CheckFlagI(argv[i], TEXT('t')))
			ctx.fDiscardStreams = TRUE;
		else if (ctx.dwAttributesMask & FILE_ATTRIBUTE_READONLY && CheckFlagI(argv[i], TEXT('r')))
			ctx.dwAttributesMask &= ~FILE_ATTRIBUTE_READONLY;
		else if (ctx.dwAttributesMask & FILE_ATTRIBUTE_HIDDEN && CheckFlagI(argv[i], TEXT('h')))
			ctx.dwAttributesMask &= ~FILE_ATTRIBUTE_HIDDEN;
		else if (ctx.dwAttributesMask & FILE_ATTRIBUTE_SYSTEM && CheckFlagI(argv[i], TEXT('s')))
			ctx.dwAttributesMask &= ~FILE_ATTRIBUTE_SYSTEM;
		else if (ctx.dwAttributesMask & FILE_ATTRIBUTE_COMPRESSED && CheckFlagI(argv[i], TEXT('c')))
			ctx.dwAttributesMask &= ~FILE_ATTRIBUTE_COMPRESSED;
		else
			fShowUsage = TRUE;
	}

	// Act according to the command line parameters

	if (fShowUsage || pszTargetPath == NULL)
	{
		static const TCHAR szUsageTemplateGeneral[] = TEXT("  -%c  %s.\n");
		static const TCHAR szUsageTemplateAttribs[] = TEXT("  -%c  Strips away all %s file attributes.\n");

		_tprintf(TEXT("Usage: vcopy [-v] [-p] [-t] [-r] [-h] [-s] [-c]\n")
		         TEXT("       [-crc32 | -md4 | -md5 | -sha1 | -ed2k] source [...] destination\n\n")
		         TEXT("Options:\n"));
		_tprintf(szUsageTemplateGeneral, TEXT('v'), TEXT("Verbose mode; prints status"));
		_tprintf(szUsageTemplateGeneral, TEXT('p'), TEXT("Preserves destination files; do not overwrite files"));
		_tprintf(szUsageTemplateGeneral, TEXT('t'), TEXT("Strips away streams and other NTFS-specific metadata"));
		_tprintf(szUsageTemplateAttribs, TEXT('r'), TEXT("read-only"));
		_tprintf(szUsageTemplateAttribs, TEXT('h'), TEXT("hidden"));
		_tprintf(szUsageTemplateAttribs, TEXT('s'), TEXT("system"));
		_tprintf(szUsageTemplateAttribs, TEXT('c'), TEXT("compressed"));
		_tprintf(TEXT("  -crc32, -md4, -md5, -sha1, or -ed2k\n")
		         TEXT("      Computes hashes; also implies -t and redirects status messages to stderr.\n"));

		return(~1);
	}
	else
	{
		/* Source->Destination behaviors:
		 * - wild->file|DNE  ERROR
		 * - wild->dir\|dir  destdir\*
		 * - file->file|DNE  destfile
		 * - file->dir\|dir  destdir\file
		 * - dir->file       ERROR
		 * - dir->dir\       destdir\dir
		 * - dir->dir|DNE    destdir
		 *
		 * Additional rules:
		 * - A destination of dir\ must be a valid existing directory
		 * - A non-wildcard source must exist
		 * - A destination may not contain a wildcard
		 */

		BOOLEAN fMustCopyIntoTarget;

		ITEMTYPE itTarget;
		itTarget.united = DetermineItemType(pszTargetPath);

		// If the user explicitly tacked on a slash to the end of the target or
		// if there are multiple source items, then the target must be a
		// directory that is copied into, not onto.

		fMustCopyIntoTarget = itTarget.fSlash || (iSourceEnd - iSourceStart > 1);

		if ( (itTarget.type != TYPE_DIR && fMustCopyIntoTarget) ||
		     (itTarget.type == TYPE_WILD) )
		{
			++ctx.cErrors;
			PrintError(pszTargetPath, (itTarget.type == TYPE_DNE) ? ERROR_PATH_NOT_FOUND : ERROR_BAD_PATHNAME);
			iSourceEnd = 0;
		}

		// When hashing, manual copy must be used, and non-error messages should
		// go to stderr to facilitate saving stdout (the hashes) to a file.
		if (ctx.fHashAlgorithm)
		{
			ctx.fDiscardStreams = TRUE;
			ctx.stream = stderr;
		}

		// Allocate a file buffer if we need to copy manually
		if (ctx.fDiscardStreams && NULL == (ctx.pvBuffer = VirtualAlloc(NULL, READ_BUFFER_SIZE, MEM_COMMIT, PAGE_READWRITE)))
			return(~0);

		for (i = iSourceStart; i < iSourceEnd; ++i)
		{
			PTSTR pszSourcePath = argv[i];

			ITEMTYPE itSource;
			itSource.united = DetermineItemType(pszSourcePath);

			if ( (itSource.type == TYPE_WILD && itTarget.type != TYPE_DIR) ||
			     (itSource.type == TYPE_DIR && itTarget.type == TYPE_FILE) )
			{
				// This will never be tripped if there are multiple sources
				++ctx.cErrors;
				PrintError(pszTargetPath, (itTarget.type == TYPE_DNE) ? ERROR_PATH_NOT_FOUND :
				                          (itTarget.type == TYPE_WILD) ? ERROR_BAD_PATHNAME : ERROR_DIRECTORY);
				continue;
			}
			else if (itSource.type == TYPE_DNE)
			{
				++ctx.cErrors;
				PrintError(pszSourcePath, (itSource.fSlash) ? ERROR_PATH_NOT_FOUND : ERROR_FILE_NOT_FOUND);
				continue;
			}
			else if (itSource.type != TYPE_DIR && itSource.fSlash)
			{
				++ctx.cErrors;
				PrintError(pszSourcePath, ERROR_DIRECTORY);
				continue;
			}

			// Dispatch the copy operation
			if (itSource.type == TYPE_FILE && itTarget.type != TYPE_DIR)
				CloneFile(&ctx, pszSourcePath, pszTargetPath, NULL);
			else if (itSource.type == TYPE_DIR && !fMustCopyIntoTarget)
				CloneDirectory(&ctx, pszSourcePath, pszTargetPath, NULL);
			else
			{
				ctx.cchTargetPrefix = SSLen(pszTargetPath) + 1;
				CloneFilesIntoDirectory(&ctx, pszSourcePath, pszTargetPath, FALSE);
			}
		}

		// Cleanup
		if (ctx.pvBuffer) VirtualFree(ctx.pvBuffer, 0, MEM_RELEASE);

		// Post-mortem
		_ftprintf(ctx.stream, TEXT("\t%d file(s) and %d folder(s) copied"), ctx.cSuccessFiles, ctx.cSuccessDirs);
		if (ctx.cSkippedFiles) _ftprintf(ctx.stream, TEXT(", %d file(s) skipped"), ctx.cSkippedFiles);
		if (ctx.cErrors) _ftprintf(ctx.stream, TEXT(", with %d error(s)"), ctx.cErrors);
		_ftprintf(ctx.stream, TEXT(".\n"));
	}

	return(ctx.cErrors);
}

VOID WINAPI CloneFile( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath, PFILEINFO pfi )
{
	BOOL fSuccess = FALSE;
	BY_HANDLE_FILE_INFORMATION bhfi;

	if (pctx->fVerbose)
		_ftprintf(pctx->stream, TEXT("%s\n"), pszSourcePath);

	/* CopyFile does more than just copy: it has special handling of reparse
	 * points, security descriptors, and NTFS streams, so until that degree of
	 * completeness can be duplicated, the manual copy can be used only when
	 * such NTFS-specific information is to be discarded; future implementation
	 * of these features will enable finer control and better efficiency.
	 */

	if (pctx->fDiscardStreams)
	{
		/* Manual copy */

		HANDLE hFileIn, hFileOut;

		hFileIn = CreateFile(
			pszSourcePath,
			GENERIC_READ,
			FILE_SHARE_READ,
			NULL,
			OPEN_EXISTING,
			FILE_FLAG_SEQUENTIAL_SCAN,
			NULL
		);

		if (hFileIn != INVALID_HANDLE_VALUE)
		{
			if (GetFileInformationByHandle(hFileIn, &bhfi))
			{
				if (pfi == NULL)
					pfi = (PFILEINFO)&bhfi;

				pfi->dwFileAttributes &= pctx->dwAttributesMask;

				// There is no need to check if the source and target files are
				// the same because the request of write access to a file that
				// does not have write sharing will fail.

				hFileOut = CreateFile(
					pszTargetPath,
					GENERIC_READ | GENERIC_WRITE | DELETE,
					FILE_SHARE_READ,
					NULL,
					(pctx->fPreserveExisting) ? CREATE_NEW : OPEN_ALWAYS,
					FILE_FLAG_SEQUENTIAL_SCAN | (pfi->dwFileAttributes & CREATION_ATTRIBUTE_MASK),
					hFileIn
				);

				if (hFileOut != INVALID_HANDLE_VALUE)
				{
					WHCTXEX whctx;

					DWORD cbRead, cbWritten;
					ULONGLONG cbTotalRead = 0, cbTotalWritten = 0;

					ULARGE_INTEGER cbFileSize;
					cbFileSize.LowPart = bhfi.nFileSizeLow;
					cbFileSize.HighPart = bhfi.nFileSizeHigh;

					// The sparse/compression attributes need to be set prior
					// to writing any data.

					SetSparseAndCompression(hFileOut, pfi->dwFileAttributes);

					// Signal the file size; this reduces disk fragmentation.

					SetFilePointer(hFileOut, bhfi.nFileSizeLow, &bhfi.nFileSizeHigh, FILE_BEGIN);

					if (SetEndOfFile(hFileOut))
					{
						SetFilePointer(hFileOut, 0, NULL, FILE_BEGIN);

						// Initiate the hashing context

						if (pctx->fHashAlgorithm)
						{
							whctx.fAlgs = pctx->fHashAlgorithm;
							WHInitEx(&whctx);
						}

						// The actual data copy...

						do
						{
							ReadFile(hFileIn, pctx->pvBuffer, READ_BUFFER_SIZE, &cbRead, NULL);

							if (cbRead)
							{
								WriteFile(hFileOut, pctx->pvBuffer, cbRead, &cbWritten, NULL);
								cbTotalRead += cbRead;
								cbTotalWritten += cbWritten;

								// Hash the data
								if (pctx->fHashAlgorithm)
									WHUpdateEx(&whctx, pctx->pvBuffer, cbRead);
							}

						} while (cbRead == READ_BUFFER_SIZE);

						if ( cbFileSize.QuadPart == cbTotalRead &&
						     cbFileSize.QuadPart == cbTotalWritten )
						{
							// Sanity check passed; set time and attributes.

							fSuccess = TRUE;
							SetTimeAndAttributes(hFileOut, pfi);

							// Push hash to stdout if the copy was successful.

							if (pctx->fHashAlgorithm)
							{
								PCTSTR pszPathToPrint = pszTargetPath + pctx->cchTargetPrefix;

								WHFinishEx(&whctx, NULL);

								switch (pctx->fHashAlgorithm)
								{
									case WHEX_CHECKCRC32:
										_tprintf(TEXT("%-12s %08x\n"), pszPathToPrint, SwapV32(*((PDWORD)whctx.results.CRC32)));
										break;
									case WHEX_CHECKMD4:
										PrintHash(whctx.results.MD4, pszPathToPrint, 4);
										break;
									case WHEX_CHECKMD5:
										PrintHash(whctx.results.MD5, pszPathToPrint, 4);
										break;
									case WHEX_CHECKSHA1:
										PrintHash(whctx.results.SHA1, pszPathToPrint, 5);
										break;
									case WHEX_CHECKED2K:
										PrintHash(whctx.results.ED2K, pszPathToPrint, 4);
										break;
								}
							}
						}
						else
						{
							// If the copied byte count isn't what was expected,
							// the preallocation size needs to be overridden.

							SetEndOfFile(hFileOut);
						}
					}
					else
					{
						// If preallocation failed, then no data will be copied,
						// and the zero-byte file should be deleted.

						IO_STATUS_BLOCK iosb;
						FILE_DISPOSITION_INFORMATION fdi = { TRUE };

						NtSetInformationFile(hFileOut, &iosb, &fdi, sizeof(fdi), FileDispositionInformation);

					} // preallocation success test

					CloseHandle(hFileOut);

				} // if valid destination handle

			} // if valid file information

			CloseHandle(hFileIn);
		}
	}
	else
	{
		/* Using CopyFile */

		if (pfi == NULL)
		{
			HANDLE hFile = CreateFile(
				pszSourcePath,
				0,
				FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
				NULL,
				OPEN_EXISTING,
				0,
				NULL
			);

			if (hFile != INVALID_HANDLE_VALUE)
			{
				if (GetFileInformationByHandle(hFile, &bhfi))
					pfi = (PFILEINFO)&bhfi;

				CloseHandle(hFile);
			}
		}

		fSuccess = CopyFile(pszSourcePath, pszTargetPath, pctx->fPreserveExisting);

		if (fSuccess && pfi)
		{
			BOOL fSetSparseOrCompression;
			HANDLE hFile;

			pfi->dwFileAttributes &= pctx->dwAttributesMask;
			fSetSparseOrCompression = pfi->dwFileAttributes & (FILE_ATTRIBUTE_SPARSE_FILE | FILE_ATTRIBUTE_COMPRESSED);

			if (fSetSparseOrCompression)
			{
				// Turn off the read-only bit so that we can modify the file.

				DWORD dwAttributes = GetFileAttributes(pszTargetPath);
				if (dwAttributes != INVALID_FILE_ATTRIBUTES && dwAttributes & FILE_ATTRIBUTE_READONLY)
					SetFileAttributes(pszTargetPath, dwAttributes & ~FILE_ATTRIBUTE_READONLY);
			}

			hFile = CreateFile(
				pszTargetPath,
				(fSetSparseOrCompression) ? GENERIC_READ | GENERIC_WRITE : FILE_WRITE_ATTRIBUTES | SYNCHRONIZE,
				FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
				NULL,
				OPEN_EXISTING,
				0,
				NULL
			);

			if (hFile != INVALID_HANDLE_VALUE)
			{
				if (fSetSparseOrCompression)
					SetSparseAndCompression(hFile, pfi->dwFileAttributes);
				SetTimeAndAttributes(hFile, pfi);
				CloseHandle(hFile);
			}
		}
	}

	if (fSuccess)
	{
		++pctx->cSuccessFiles;
	}
	else
	{
		DWORD dwErrorCode = GetLastError();

		if (dwErrorCode == ERROR_FILE_EXISTS && pctx->fPreserveExisting)
		{
			++pctx->cSkippedFiles;
			if (pctx->fVerbose)
				_ftprintf(stderr, TEXT("Already exists: %s\n"), pszTargetPath);
		}
		else
		{
			++pctx->cErrors;
			PrintError(pszSourcePath, dwErrorCode);
		}
	}
}

VOID WINAPI CloneDirectory( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath, PFILEINFO pfi )
{
	BY_HANDLE_FILE_INFORMATION bhfi;

	if (pctx->fVerbose)
		_ftprintf(pctx->stream, TEXT("%s\n"), pszSourcePath);

	// Step 1: Create the target directory

	(pctx->fDiscardStreams) ?
		CreateDirectory(pszTargetPath, NULL) :
		CreateDirectoryEx(pszSourcePath, pszTargetPath, NULL);

	// Step 2: Clone the attributes

	if (IsValidDirectory(pszTargetPath))
	{
		// Ensure that we have the source attribute information

		if (pfi == NULL)
		{
			HANDLE hFile = CreateFile(
				pszSourcePath,
				0,
				FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
				NULL,
				OPEN_EXISTING,
				FILE_FLAG_BACKUP_SEMANTICS,
				NULL
			);

			if (hFile != INVALID_HANDLE_VALUE)
			{
				if (GetFileInformationByHandle(hFile, &bhfi))
					pfi = (PFILEINFO)&bhfi;

				CloseHandle(hFile);
			}
		}

		// If the folder is compressed or encrypted, we should set the right
		// bits BEFORE copying any files...

		if (pfi)
		{
			pfi->dwFileAttributes &= pctx->dwAttributesMask;

			if (pfi->dwFileAttributes & FILE_ATTRIBUTE_COMPRESSED)
			{
				HANDLE hFile = CreateFile(
					pszTargetPath,
					GENERIC_READ | GENERIC_WRITE,
					FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
					NULL,
					OPEN_EXISTING,
					FILE_FLAG_BACKUP_SEMANTICS,
					NULL
				);

				if (hFile != INVALID_HANDLE_VALUE)
				{
					SetSparseAndCompression(hFile, FILE_ATTRIBUTE_COMPRESSED);
					CloseHandle(hFile);
				}
			}
			else if (pfi->dwFileAttributes & FILE_ATTRIBUTE_ENCRYPTED)
			{
				EncryptFile(pszTargetPath);
			}
		}
	}
	else
	{
		++pctx->cErrors;
		PrintError(pszTargetPath, GetLastError());
		return;
	}

	// Step 3: Clone the content

	CloneFilesIntoDirectory(pctx, pszSourcePath, pszTargetPath, TRUE);

	// Step 4: Clone the timestamp

	if (pfi)
	{
		HANDLE hFile = CreateFile(
			pszTargetPath,
			FILE_WRITE_ATTRIBUTES | SYNCHRONIZE,
			FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
			NULL,
			OPEN_EXISTING,
			FILE_FLAG_BACKUP_SEMANTICS,
			NULL
		);

		if (hFile != INVALID_HANDLE_VALUE)
		{
			SetTimeAndAttributes(hFile, pfi);
			CloseHandle(hFile);
		}
	}

	++pctx->cSuccessDirs;
}

VOID WINAPI CloneFilesIntoDirectory( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath, BOOLEAN fSourceIsDirectoryRoot )
{
	HANDLE hFind;
	WIN32_FIND_DATA finddata;

	// Copy the paths to a scratchpad buffer
	TCHAR szSourcePath[MAX_PATH << 2];
	TCHAR szTargetPath[MAX_PATH << 2];
	PTSTR pszSourcePathAppend = SSChainNCpy(szSourcePath, pszSourcePath, SSLen(pszSourcePath));
	PTSTR pszTargetPathAppend = SSChainNCpy(szTargetPath, pszTargetPath, SSLen(pszTargetPath));

	// Ensure that we have a slash termination for the target path
	*pszTargetPathAppend++ = TEXT('\\');

	// Prep parameters for FindFirstFile and the establish the append position
	*pszSourcePathAppend = 0;
	if (fSourceIsDirectoryRoot)
		SSCpy4Ch(pszSourcePathAppend++, TEXT('\\'), TEXT('*'), 0, 0);
	else if (pszSourcePathAppend = StrRChr(szSourcePath, pszSourcePathAppend, TEXT('\\')))
		++pszSourcePathAppend;
	else
		pszSourcePathAppend = szSourcePath;

	if ((hFind = FindFirstFile(szSourcePath, &finddata)) != INVALID_HANDLE_VALUE)
	{
		do
		{
			if (!IsSpecialDirectoryName(finddata.cFileName))
			{
				SSCpy(pszSourcePathAppend, finddata.cFileName);
				SSCpy(pszTargetPathAppend, finddata.cFileName);

				if (finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
					CloneDirectory(pctx, szSourcePath, szTargetPath, (PFILEINFO)&finddata);
				else
					CloneFile(pctx, szSourcePath, szTargetPath, (PFILEINFO)&finddata);
			}

		} while (FindNextFile(hFind, &finddata));

		FindClose(hFind);
	}
}

VOID WINAPI PrintHash( PVOID pvResult, PCTSTR pszPath, SIZE_T cdwResult )
{
	PDWORD pdwResult = pvResult;

	while (cdwResult)
	{
		_tprintf(TEXT("%08x"), SwapV32(*pdwResult));
		++pdwResult;
		--cdwResult;
	}

	_tprintf(TEXT(" *%s\n"), pszPath);
}

UINT16 __fastcall DetermineItemType( PTSTR pszPath )
{
	PTSTR psz = pszPath;

	ITEMTYPE it;
	it.united = 0;

	while (*psz)
	{
		switch (*psz)
		{
			case TEXT('*'):
			case TEXT('?'):
				it.type = TYPE_WILD;
				break;

			case TEXT('/'):
				*psz = TEXT('\\');

			// Any encounter with a slash following a wildcard must be invalid,
			// and we flag that by setting fSlash with TYPE_WILD.
			case TEXT('\\'):
				if (it.type == TYPE_WILD)
					it.fSlash = TRUE;
				break;
		}

		++psz;
	}

	while (--psz >= pszPath && *psz == TEXT('\\'))
	{
		*psz = 0;
		it.fSlash = TRUE;
	}

	if (it.fSlash && *pszPath == 0)
	{
		// Special case: "\"
		it.type = TYPE_DIR;
	}
	else if (it.type == TYPE_DNE)
	{
		DWORD dwAttributes = GetFileAttributes(pszPath);

		if (dwAttributes != INVALID_FILE_ATTRIBUTES)
			it.type = (dwAttributes & FILE_ATTRIBUTE_DIRECTORY) ? TYPE_DIR : TYPE_FILE;
	}

	return(it.united);
}

VOID __fastcall SetSparseAndCompression( HANDLE hFile, DWORD dwAttributes )
{
	DWORD cbRet;

	if (dwAttributes & FILE_ATTRIBUTE_SPARSE_FILE)
	{
		FILE_SET_SPARSE_BUFFER buf = { TRUE };
		DeviceIoControl(hFile, FSCTL_SET_SPARSE, &buf, sizeof(buf), NULL, 0, &cbRet, NULL);
	}

	if (dwAttributes & FILE_ATTRIBUTE_COMPRESSED)
	{
		USHORT buf = COMPRESSION_FORMAT_DEFAULT;
		DeviceIoControl(hFile, FSCTL_SET_COMPRESSION, &buf, sizeof(buf), NULL, 0, &cbRet, NULL);
	}
}

VOID __fastcall SetTimeAndAttributes( HANDLE hFile, PFILEINFO pfi )
{
	IO_STATUS_BLOCK iosb;
	FILE_BASIC_INFORMATION fbi;

	memcpy(&fbi, &pfi->ftCreationTime, sizeof(FILETIME) * 3);
	fbi.ChangeTime.QuadPart = 0;
	fbi.FileAttributes = pfi->dwFileAttributes;

	NtSetInformationFile(hFile, &iosb, &fbi, sizeof(fbi), FileBasicInformation);
}

VOID __fastcall PrintError( PCTSTR pszPath, DWORD dwErrorCode )
{
	_ftprintf(stderr, TEXT("Error: %s"), pszPath);

	if (dwErrorCode)
	{
		PTSTR pszErrorMessage = NULL;

		DWORD cchMessage = FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
			NULL,
			dwErrorCode,
			0,
			(PTSTR)&pszErrorMessage,
			0,
			NULL
		);

		if (cchMessage)
		{
			pszErrorMessage[StrCSpn(pszErrorMessage, TEXT("\r\n"))] = 0;
			_ftprintf(stderr, TEXT(" [%s]"), pszErrorMessage);
		}
		else
		{
			_ftprintf(stderr, TEXT(" [%08X]"), dwErrorCode);
		}

		LocalFree(pszErrorMessage);
	}

	_ftprintf(stderr, TEXT("\n"));
}

__forceinline BOOLEAN IsFlag( PCTSTR pszArg )
{
	return(
		(pszArg[0] | 0x02) == TEXT('/') &&
		(pszArg[1]       ) != 0 &&
		(pszArg[2]       ) == 0
	);
}

__forceinline BOOLEAN CheckFlagI( PCTSTR pszArg, TCHAR ch )
{
	return(
		(pszArg[1] | 0x20) == ch
	);
}

__forceinline BOOLEAN IsSpecialDirectoryName( PCTSTR pszPath )
{
	// TRUE if name is "." or ".."

	#ifdef UNICODE
	return(
		(*((UPDWORD)pszPath) == WCHARS2DWORD(L'.', 0)) ||
		(*((UPDWORD)pszPath) == WCHARS2DWORD(L'.', L'.') && pszPath[2] == 0)
	);
	#else
	return(
		(*((UPWORD)pszPath) == CHARS2WORD('.', 0)) ||
		(*((UPWORD)pszPath) == CHARS2WORD('.', '.') && pszPath[2] == 0)
	);
	#endif
}

__forceinline BOOLEAN IsValidDirectory( PCTSTR pszPath )
{
	DWORD dwAttributes = GetFileAttributes(pszPath);
	return(dwAttributes != INVALID_FILE_ATTRIBUTES && dwAttributes & FILE_ATTRIBUTE_DIRECTORY);
}
